package ir.sharif.ce.partov.user;

import java.util.ArrayList;
import java.util.Scanner;
import java.util.StringTokenizer;

import ir.sharif.ce.partov.base.ClientFramework;
import ir.sharif.ce.partov.base.Frame;
import ir.sharif.ce.partov.base.Machine;
import ir.sharif.ce.partov.utils.Utility;

public class SimulateMachine extends Machine {

	// You will need this to know which node you are
	private int index = 0;

	public static final int GPG_HEADER_LENGTH = 2 ;

	private ArrayList<RoutingTableEntry> routing_table = new ArrayList<>() ;

	public SimulateMachine(ClientFramework clientFramework, int count) {
		super(clientFramework, count);
		// The machine instantiated.
		// Interfaces are not valid at this point.
	}

	public void initialize() {
		// TODO: Initialize your program here; interfaces are valid now.
		parseRoutingTables();
	}

	/**
	 * This method is called from the main thread. Also ownership of the data of
	 * the frame is not with you. If you need it, make a copy for yourself.
	 */
	public void processFrame(Frame frame, int ifaceIndex) {
		// TODO: process the raw frame; frame.data points to the frame byte

		//make a copy of frame data
		byte[] data = new byte[frame.data.length];
		System.arraycopy(frame.data, 0, data, 0, data.length);
		
		//simple frame data processing sample
		System.out.println("Frame received at iface " + ifaceIndex
		+ " with length " + frame.length );
		EthernetHeader eth = new EthernetHeader(data, 0);
		System.out.println("type of ethernet packet is 0x"+ Utility.byteToHex(eth.getTypeInBytes()[0])+Utility.byteToHex(eth.getTypeInBytes()[1]));
		if(eth.getTypeAsInt()==((int)IPv4Header.IP_PROTOCOL)){
			IPv4Header iph = new IPv4Header(data, 14, 5);	//Packet, Position of IP header in packet, IP header Length in Word unit(4 bytes)
			System.out.println("Source IP: "+ Utility.getIPString(iph.getSrc()));
			System.out.println("Destination IP: "+ Utility.getIPString(iph.getDest()));
		}
	}

	void sendPacketToIndex(GPGHeader gpg_hdr_tmp, byte[] payload_tmp, int payload_len, int link) {

		int packet_size = EthernetHeader.ETHERNET_LENGTH +
						(IPv4Header.LEAST_IHL * IPv4Header.WORD_SIZE) +
						UDP.Default_len + GPG_HEADER_LENGTH + payload_len ;

		byte[] packet = new byte[packet_size] ;

		RoutingTableEntry entry = getRoutingInformation(link) ;

		if (entry == null) {
			return ;
		}

		EthernetHeader ethernet_hdr = new EthernetHeader(packet, 0) ;
		ethernet_hdr.setType((int)IPv4Header.IP_PROTOCOL);
		ethernet_hdr.setSrc(iface[link].mac) ;
		ethernet_hdr.setDest(entry.mac) ;

		IPv4Header ip_hdr = new IPv4Header(packet, EthernetHeader.ETHERNET_LENGTH, IPv4Header.LEAST_IHL) ;
		ip_hdr.setDefaults() ;
		ip_hdr.setTotalLength(IPv4Header.LEAST_IHL * IPv4Header.WORD_SIZE + UDP.Default_len + GPG_HEADER_LENGTH + payload_len) ;
		ip_hdr.setSrc(iface[link].getIp()) ;
		ip_hdr.setDest(entry.ip) ;

		UDP udp_hdr = new UDP(packet, EthernetHeader.ETHERNET_LENGTH + IPv4Header.LEAST_IHL * IPv4Header.WORD_SIZE, -1) ;
		udp_hdr.setDefaults() ;
		udp_hdr.setSrcPort(3000) ;
		udp_hdr.setDestPort(3000) ;
		udp_hdr.setLen(GPG_HEADER_LENGTH + payload_len) ;

		GPGHeader gpg_hdr = new GPGHeader(packet, EthernetHeader.ETHERNET_LENGTH + IPv4Header.LEAST_IHL * IPv4Header.WORD_SIZE + UDP.Default_len) ;
		System.arraycopy(gpg_hdr_tmp.data, 0, gpg_hdr.data, 0, 2);

		if (payload_len != 0) {
			System.arraycopy(payload_tmp, 0, packet, EthernetHeader.ETHERNET_LENGTH + IPv4Header.LEAST_IHL * IPv4Header.WORD_SIZE +
					UDP.Default_len + GPG_HEADER_LENGTH, payload_len);
		}

		Frame out_frame = new Frame(packet_size, packet);
		sendFrame (out_frame, link);
	}

	RoutingTableEntry getRoutingInformation(int index) {
		for (int i = 0 ; i < routing_table.size() ; i++) {
			if (routing_table.get(i).index == index) {
				return routing_table.get(i) ;
			}
		}
		return null ;
	}

	/**
	 * This method will be run from an independent thread. Use it if needed or
	 * simply return. Returning from this method will not finish the execution
	 * of the program.
	 */
	public void run() {
		// TODO: write you business logic here...
		// TODO: remove the following code, its just a tutorial

		Scanner s = new Scanner(System.in);
		System.out.println("Type \"send\" to send a sample frame.");
		while(!s.nextLine().equals("send"));
		
		EthernetHeader eth = new EthernetHeader(EthernetHeader.BROADCAST, iface[0].mac, IPv4Header.IP_PROTOCOL);
		IPv4Header iph = new IPv4Header();
		iph.setTotalLength(100);
		iph.setSrc(iface[0].ip);
		iph.setDest(IPv4Header.BROADCAST_IP);
		iph.setTTL(64);
		
		byte[] data = new byte[14+100];
		System.arraycopy(eth.getData(), 0, data, 0, 14);
		System.arraycopy(iph.getData(), 0, data, 14, iph.getData().length);
		this.sendFrame(new Frame(114,data), 0);
		System.out.println("Sample frame sent.");
		s.close();
	}

	public void parseRoutingTables() {
		String data = getCustomInformation() ;
		ArrayList<String> lines = new ArrayList<>() ;
		StringTokenizer tokenizer = new StringTokenizer(data, "\n") ;
		while(tokenizer.hasMoreElements()) {
			lines.add(tokenizer.nextToken()) ;
		}
		for(int i = 0 ; i < lines.size() ; i++){
			System.err.println("#" + i + ": " + lines.get(i)) ;
		}
		this.index = 0 ;
		for(int i = 2 ; i < lines.size() ; i++){
			String line = lines.get(i) ;
			// ==================================================================== //
			RoutingTableEntry entry = new RoutingTableEntry() ;
			//get IP
			tokenizer = new StringTokenizer(line) ;
			String token = tokenizer.nextToken() ;
			entry.index = Integer.parseInt(token) ;
			if (this.index == entry.index) {
				this.index ++ ;
			}
			//get network mask
			token = tokenizer.nextToken() ;
			entry.mask = Utility.getIP(token) ;
			//get next hop IP
			token = tokenizer.nextToken() ;
			entry.ip = Utility.getIP(token) ;
			//get index of egress interface
			token = tokenizer.nextToken() ;
			entry.link = Integer.parseInt(token) ;
			//get mac address
			token = tokenizer.nextToken() ;
			entry.mac = Utility.getMac(token) ;
			//save entry
			routing_table.add(entry);
			entry.print();
		}
		System.err.println("Index" + this.index) ;
	}

	class RoutingTableEntry {
		int index ;
		int mask ;
		int link ;
		int ip ;
		byte[] mac = new byte[6] ;

		public void print() {
			System.err.println("Index: " + index) ;
			System.err.println("Mask: " + Utility.getIPString(mask)) ;
			System.err.println("IP: " + Utility.getIPString(ip)) ;
			System.err.println("Interface: " + link) ;
		}
	}

	class GPGHeader {
		public byte[] data;
		
		public GPGHeader(byte[] packet, int pos) {
			data = new byte[2];
			System.arraycopy(packet, pos, data, 0, 2);
		}
	}

	int generatePrime(int k, int e, boolean odd) {
		int begin = 1 << (k - 1) ; // k bits
		int end = 1 << k ; // k + 1 bits
		int n = 0 ;
		for (int i = begin ; i < end ; i++) {
			//cerr << "testing " << i << endl ;
			if (isPrime(i) && i % e != 1) {
				n ++ ;
				if (odd) {
					if (n == this.index) {
						return i ;
					}
				} else {
					if (n == 5 - this.index) {
						return i ;
					}
				}
			}
		}
		return -1 ;
	}

	boolean isPrime(int a) {
		//cerr << "testing primality for " << a << endl ;
		for(int j = 2 ; j < a ; j++) {
			if(a % j == 0) {
				//cerr << "not prime" << endl ;
				return false ;
			}
		}
		//cerr << "prime" << endl ;
		return true ;
	}

	int modularInverse(int u, int v) {
		int inv, u1, u3, v1, v3, t1, t3, q;
		int iter;
	    /* Step X1. Initialise */
		u1 = 1;
		u3 = u;
		v1 = 0;
		v3 = v;
	    /* Remember odd/even iterations */
		iter = 1;
	    /* Step X2. Loop while v3 != 0 */
		while (v3 != 0)
		{
	        /* Step X3. Divide and "Subtract" */
			q = u3 / v3;
			t3 = u3 % v3;
			t1 = u1 + q * v1;
	        /* Swap */
			u1 = v1; v1 = t1; u3 = v3; v3 = t3;
			iter = -iter;
		}
	    /* Make sure u3 = gcd(u,v) == 1 */
		if (u3 != 1)
			return 0;   /* Error: No inverse exists */
	    /* Ensure a positive result */
		if (iter < 0)
			inv = v - u1;
		else
			inv = u1;
		return inv;
	}

}